@@ -12,6 +12,8 @@ fi
 
 options="-n --exclude=moc_* --exclude=qrc_* --include=*.cpp --include=*.h -r"
 
+grep $options '( ' $args
+grep $options ' )' $args
 grep $options 'if(' $args
 grep $options 'for(' $args
 grep $options 'while(' $args
@@ -22,6 +24,7 @@ grep $options '^{ }*{' $args
 grep $options '){' $args
 grep $options '	' $args
 grep $options ' \*>' $args
+grep $options '#include "q' $args
 egrep $options '\(.*\* .*\)' $args | grep '::'
 grep $options '[a-z]++' $args | grep 'for'
 grep $options '[a-z]--' $args | grep 'for'
