@@ -1,5 +1,5 @@
 /*
- * Copyright 2008 Benjamin C. Meyer <ben@meyerhome.net>
+ * Copyright 2008-2009 Benjamin C. Meyer <ben@meyerhome.net>
  * Copyright 2008 Jason A. Donenfeld <Jason@zx2c4.com>
  * Copyright 2008 Ariya Hidayat <ariya.hidayat@gmail.com>
  *
@@ -64,6 +64,10 @@
 
 #include "webview.h"
 
+#include "adblockdialog.h"
+#include "adblockmanager.h"
+#include "adblockpage.h"
+#include "autofillmanager.h"
 #include "addbookmarkdialog.h"
 #include "bookmarksmanager.h"
 #include "browserapplication.h"
@@ -79,6 +83,7 @@
 #include <qdebug.h>
 #include <qevent.h>
 #include <qmenubar.h>
+#include <qtimer.h>
 #include <qwebframe.h>
 
 #if QT_VERSION >= 0x040600 || defined(WEBKIT_TRUNK)
@@ -89,6 +94,7 @@ Q_DECLARE_METATYPE(QWebElement)
 #include <qsettings.h>
 #include <qtooltip.h>
 #include <qwebelement.h>
+#include <qwindowsstyle.h>
 #endif
 
 #include <qdebug.h>
@@ -98,12 +104,20 @@ WebView::WebView(QWidget *parent)
     , m_progress(0)
     , m_currentZoom(100)
     , m_page(new WebPage(this))
-#ifdef WEBKIT_TRUNK
+#if QT_VERSION >= 0x040600 || defined(WEBKIT_TRUNK)
     , m_enableAccessKeys(true)
     , m_accessKeysPressed(false)
 #endif
 {
     setPage(m_page);
+#if QT_VERSION >= 0x040600
+    QPalette p;
+    if (p.color(QPalette::Window) != Qt::white) {
+        QWindowsStyle s;
+        p = s.standardPalette();
+        setPalette(p);
+    }
+#endif
     connect(page(), SIGNAL(statusBarMessage(const QString&)),
             SLOT(setStatusBarText(const QString&)));
     connect(this, SIGNAL(loadProgress(int)),
@@ -114,10 +128,8 @@ WebView::WebView(QWidget *parent)
             this, SIGNAL(urlChanged(const QUrl &)));
     connect(page(), SIGNAL(downloadRequested(const QNetworkRequest &)),
             this, SLOT(downloadRequested(const QNetworkRequest &)));
-#if QT_VERSION >= 0x040500
     connect(BrowserApplication::instance(), SIGNAL(zoomTextOnlyChanged(bool)),
             this, SLOT(applyZoom()));
-#endif
     page()->setForwardUnsupportedContent(true);
     setAcceptDrops(true);
 
@@ -131,6 +143,7 @@ WebView::WebView(QWidget *parent)
     connect(m_page, SIGNAL(scrollRequested(int, int, const QRect &)),
             this, SLOT(hideAccessKeys()));
 #endif
+    loadSettings();
 }
 
 void WebView::loadSettings()
@@ -145,6 +158,11 @@ void WebView::loadSettings()
 
 #if !(QT_VERSION >= 0x040600 || defined(WEBKIT_TRUNK))
 #include <qdir.h>
+// DO NOT CHANGE ANYTHING IN THIS FUNCTION
+// You want to change TabWidget::guessUrlFromString()
+// This is a copy of QWebView::guessUrlFromStringis from
+// QtWebKit and should never be out of sync with that code.
+// This function is fragile, very easy to break, and tested in QtWebKit.
 QUrl WebView::guessUrlFromString(const QString &string)
 {
     QString trimmedString = string.trimmed();
@@ -207,8 +225,6 @@ void WebView::contextMenuEvent(QContextMenuEvent *event)
         if (!page()->selectedText().isEmpty())
             menu->addAction(pageAction(QWebPage::Copy));
         menu->addAction(tr("&Copy Link Location"), this, SLOT(copyLinkToClipboard()));
-        if (page()->settings()->testAttribute(QWebSettings::DeveloperExtrasEnabled))
-            menu->addAction(pageAction(QWebPage::InspectElement));
     }
 
     if (!r.imageUrl().isEmpty()) {
@@ -222,6 +238,8 @@ void WebView::contextMenuEvent(QContextMenuEvent *event)
         menu->addAction(tr("&Save Image"), this, SLOT(downloadImageToDisk()));
         menu->addAction(tr("&Copy Image"), this, SLOT(copyImageToClipboard()));
         menu->addAction(tr("C&opy Image Location"), this, SLOT(copyImageLocationToClipboard()))->setData(r.imageUrl().toString());
+        menu->addSeparator();
+        menu->addAction(tr("Block Image"), this, SLOT(blockImage()))->setData(r.imageUrl().toString());
     }
 
     if (!page()->selectedText().isEmpty()) {
@@ -261,12 +279,13 @@ void WebView::contextMenuEvent(QContextMenuEvent *event)
     }
 #endif
 
-#if QT_VERSION >= 0x040500
     if (menu->isEmpty()) {
         delete menu;
         menu = page()->createStandardContextMenu();
+    } else {
+        if (page()->settings()->testAttribute(QWebSettings::DeveloperExtrasEnabled))
+            menu->addAction(pageAction(QWebPage::InspectElement));
     }
-#endif
 
     if (!menu->isEmpty()) {
         if (BrowserMainWindow::parentWindow(tabWidget())->menuBar()->isHidden()) {
@@ -358,6 +377,15 @@ void WebView::copyImageLocationToClipboard()
     }
 }
 
+void WebView::blockImage()
+{
+    if (QAction *action = qobject_cast<QAction*>(sender())) {
+        QString imageUrl = action->data().toString();
+        AdBlockDialog *dialog = AdBlockManager::instance()->showDialog();
+        dialog->addCustomRule(imageUrl);
+    }
+}
+
 void WebView::bookmarkLink()
 {
     if (QAction *action = qobject_cast<QAction*>(sender())) {
@@ -409,13 +437,13 @@ void WebView::addSearchEngine()
         return;
     }
 
-    QUrl searchUrl(url().resolved(QUrl(formElement.attribute(QLatin1String("action")))));
+    QUrl searchUrl(page()->mainFrame()->baseUrl().resolved(QUrl(formElement.attribute(QLatin1String("action")))));
     QMap<QString, QString> searchEngines;
     QList<QWebElement> inputFields = formElement.findAll(QLatin1String("input"));
-    foreach (const QWebElement &inputField, inputFields) {
+    foreach (QWebElement inputField, inputFields) {
         QString type = inputField.attribute(QLatin1String("type"), QLatin1String("text"));
         QString name = inputField.attribute(QLatin1String("name"));
-        QString value = inputField.scriptableProperty(QLatin1String("value")).toString();
+        QString value = inputField.evaluateJavaScript(QLatin1String("this.value")).toString();
 
         if (type == QLatin1String("submit")) {
             searchEngines.insert(value, name);
@@ -425,7 +453,7 @@ void WebView::addSearchEngine()
 
             searchUrl.addQueryItem(name, value);
         } else if (type == QLatin1String("checkbox") || type == QLatin1String("radio")) {
-            if (inputField.scriptableProperty(QLatin1String("checked")).toBool()) {
+            if (inputField.evaluateJavaScript(QLatin1String("this.checked")).toBool()) {
                 searchUrl.addQueryItem(name, value);
             }
         } else if (type == QLatin1String("hidden")) {
@@ -434,9 +462,9 @@ void WebView::addSearchEngine()
     }
 
     QList<QWebElement> selectFields = formElement.findAll(QLatin1String("select"));
-    foreach (const QWebElement &selectField, selectFields) {
+    foreach (QWebElement selectField, selectFields) {
         QString name = selectField.attribute(QLatin1String("name"));
-        int selectedIndex = selectField.scriptableProperty(QLatin1String("selectedIndex")).toInt();
+        int selectedIndex = selectField.evaluateJavaScript(QLatin1String("this.selectedIndex")).toInt();
         if (selectedIndex == -1)
             continue;
 
@@ -506,11 +534,7 @@ int WebView::levelForZoom(int zoom)
 
 void WebView::applyZoom()
 {
-#if QT_VERSION >= 0x040500
     setZoomFactor(qreal(m_currentZoom) / 100.0);
-#else
-    setTextSizeMultiplier(qreal(m_currentZoom) / 100.0);
-#endif
 }
 
 void WebView::zoomIn()
@@ -544,17 +568,15 @@ void WebView::loadFinished()
                    << "Url:" << url();
     }
     m_progress = 0;
+    AdBlockManager::instance()->page()->applyRulesToPage(page());
+    BrowserApplication::instance()->autoFillManager()->fill(page());
 }
 
 void WebView::loadUrl(const QUrl &url, const QString &title)
 {
     if (url.scheme() == QLatin1String("javascript")) {
-        QString scriptSource = url.toString().mid(11);
+        QString scriptSource = QUrl::fromPercentEncoding(url.toString(Q_FLAGS(QUrl::TolerantMode|QUrl::RemoveScheme)).toAscii());
         QVariant result = page()->mainFrame()->evaluateJavaScript(scriptSource);
-        if (result.canConvert(QVariant::String)) {
-            QString newHtml = result.toString();
-            setHtml(newHtml);
-        }
         return;
     }
     m_initialUrl = url;
@@ -604,21 +626,26 @@ void WebView::dragEnterEvent(QDragEnterEvent *event)
 
 void WebView::dragMoveEvent(QDragMoveEvent *event)
 {
-    if (!event->mimeData()->urls().isEmpty()) {
-        event->acceptProposedAction();
-    } else {
-        QUrl url(event->mimeData()->text());
-        if (url.isValid())
+    event->ignore();
+    if (event->source() != this) {
+        if (!event->mimeData()->urls().isEmpty()) {
             event->acceptProposedAction();
+        } else {
+            QUrl url(event->mimeData()->text());
+            if (url.isValid())
+                event->acceptProposedAction();
+        }
     }
-    if (!event->isAccepted())
+    if (!event->isAccepted()) {
         QWebView::dragMoveEvent(event);
+    }
 }
 
 void WebView::dropEvent(QDropEvent *event)
 {
     QWebView::dropEvent(event);
     if (!event->isAccepted()
+        && event->source() != this
         && event->possibleActions() & Qt::CopyAction) {
 
         QUrl url;
@@ -664,7 +691,6 @@ void WebView::keyPressEvent(QKeyEvent *event)
     if (m_enableAccessKeys) {
         m_accessKeysPressed = (event->modifiers() == Qt::ControlModifier
                                && event->key() == Qt::Key_Control);
-
         if (!m_accessKeysPressed) {
             if (checkForAccessKey(event)) {
                 hideAccessKeys();
@@ -672,9 +698,13 @@ void WebView::keyPressEvent(QKeyEvent *event)
                 return;
             }
             hideAccessKeys();
+        } else {
+            QTimer::singleShot(200, this, SLOT(accessKeyShortcut()));
         }
     }
 #endif
+
+#if QT_VERSION < 0x040600
     switch (event->key()) {
     case Qt::Key_Back:
         pageAction(WebPage::Back)->trigger();
@@ -694,22 +724,41 @@ void WebView::keyPressEvent(QKeyEvent *event)
         break;
     default:
         QWebView::keyPressEvent(event);
+        return;
     }
+#endif
+    QWebView::keyPressEvent(event);
 }
 
 #if QT_VERSION >= 0x040600 || defined(WEBKIT_TRUNK)
+void WebView::accessKeyShortcut()
+{
+    if (!hasFocus()
+        || !m_accessKeysPressed
+        || !m_enableAccessKeys)
+        return;
+    if (m_accessKeyLabels.isEmpty()) {
+        showAccessKeys();
+    } else {
+        hideAccessKeys();
+    }
+    m_accessKeysPressed = false;
+}
+
 void WebView::keyReleaseEvent(QKeyEvent *event)
+{
+    if (m_enableAccessKeys)
+        m_accessKeysPressed = event->key() == Qt::Key_Control;
+    QWebView::keyReleaseEvent(event);
+}
+
+void WebView::focusOutEvent(QFocusEvent *event)
 {
     if (m_accessKeysPressed) {
-        if (m_accessKeyLabels.isEmpty()) {
-            showAccessKeys();
-        } else {
-            hideAccessKeys();
-        }
+        hideAccessKeys();
         m_accessKeysPressed = false;
     }
-
-    QWebView::keyReleaseEvent(event);
+    QWebView::focusOutEvent(event);
 }
 
 bool WebView::checkForAccessKey(QKeyEvent *event)
